@@ -77,7 +77,7 @@ extern int libconfig_yylex_init_extra(struct scan_context *scan_ctx,
 
 /* ------------------------------------------------------------------------- */
 
-static void __config_locale_override()
+static void __config_locale_override(void)
 {
 #if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__)) \
   && ! defined(__MINGW32__)
@@ -104,7 +104,7 @@ static void __config_locale_override()
 
 /* ------------------------------------------------------------------------- */
 
-static void __config_locale_restore()
+static void __config_locale_restore(void)
 {
 #if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__)) \
   && ! defined(__MINGW32__)
@@ -304,7 +304,8 @@ static void __config_write_value(const config_value_t *value, int type,
 
         for(s = list->elements; len--; s++)
         {
-          __config_write_value(&((*s)->value), (*s)->type, (*s)->format,
+          __config_write_value(&((*s)->value), (*s)->type,
+                               config_setting_get_format(*s),
                                depth + 1, tab_width, stream);
 
           if(len)
@@ -332,7 +333,8 @@ static void __config_write_value(const config_value_t *value, int type,
 
         for(s = list->elements; len--; s++)
         {
-          __config_write_value(&((*s)->value), (*s)->type, (*s)->format,
+          __config_write_value(&((*s)->value), (*s)->type,
+                               config_setting_get_format(*s),
                                depth + 1, tab_width, stream);
 
           if(len)
@@ -557,13 +559,16 @@ static int __config_read(config_t *config, FILE *stream, const char *filename,
 
   /* Reinitialize the config */
   void (*destructor)(void *) = config->destructor;
+  const char *include_dir = config->include_dir;
   unsigned short tab_width = config->tab_width;
   unsigned short flags = config->flags;
 
+  config->include_dir = NULL;
   config_destroy(config);
   config_init(config);
 
   config->destructor = destructor;
+  config->include_dir = include_dir;
   config->tab_width = tab_width;
   config->flags = flags;
 
@@ -583,12 +588,19 @@ static int __config_read(config_t *config, FILE *stream, const char *filename,
   else /* read from string */
     buffer = libconfig_yy_scan_string(str, scanner);
 
+  libconfig_yyset_lineno(1, scanner);
   r = libconfig_yyparse(scanner, &parse_ctx, &scan_ctx);
 
   if(r != 0)
   {
+    YY_BUFFER_STATE buf;
+
     config->error_file = scanctx_current_filename(&scan_ctx);
     config->error_type = CONFIG_ERR_PARSE;
+
+    /* Unwind the include stack, freeing the buffers and closing the files. */
+    while((buf = (YY_BUFFER_STATE)scanctx_pop_include(&scan_ctx)) != NULL)
+      libconfig_yy_delete_buffer(buf, scanner);
   }
 
   libconfig_yylex_destroy(scanner);
@@ -629,7 +641,8 @@ static void __config_write_setting(const config_setting_t *setting,
     fprintf(stream, " %c ", (setting->type == CONFIG_TYPE_GROUP ? ':' : '='));
   }
 
-  __config_write_value(&(setting->value), setting->type, setting->format,
+  __config_write_value(&(setting->value), setting->type,
+                       config_setting_get_format(setting),
                        depth, tab_width, stream);
 
   if(depth > 0)
@@ -698,7 +711,9 @@ void config_destroy(config_t *config)
 
   for(f = config->filenames; count > 0; ++f, --count)
     _delete(*f);
+
   _delete(config->filenames);
+  _delete(config->include_dir);
 
   memset((void *)config, 0, sizeof(config_t));
 }
@@ -765,28 +780,72 @@ static config_setting_t *config_setting_create(config_setting_t *parent,
 
 /* ------------------------------------------------------------------------- */
 
-int config_setting_get_int(const config_setting_t *setting)
+static int __config_setting_get_int(const config_setting_t *setting,
+                                    int *value)
 {
   switch(setting->type)
   {
     case CONFIG_TYPE_INT:
-      return(setting->value.ival);
+      *value = setting->value.ival;
+      return(CONFIG_TRUE);
 
     case CONFIG_TYPE_INT64:
       if((setting->value.llval > INT32_MAX)
          || (setting->value.llval < INT32_MIN))
-        return(0);
+        *value = 0;
       else
-        return((int)setting->value.llval);
+        *value = (int)(setting->value.llval);
+      return(CONFIG_TRUE);
 
     case CONFIG_TYPE_FLOAT:
       if((setting->config->flags & CONFIG_OPTION_AUTOCONVERT) != 0)
-        return((int)(setting->value.fval));
+      {
+        *value = (int)(setting->value.fval);
+        return(CONFIG_TRUE);
+      }
       else
-        /* fall through */;
+      { /* fall through */ }
 
     default:
-      return(0);
+      return(CONFIG_FALSE);
+  }
+}
+
+/* ------------------------------------------------------------------------- */
+
+int config_setting_get_int(const config_setting_t *setting)
+{
+  int value = 0;
+  __config_setting_get_int(setting, &value);
+  return(value);
+}
+
+/* ------------------------------------------------------------------------- */
+
+static int __config_setting_get_int64(const config_setting_t *setting,
+                                      long long *value)
+{
+  switch(setting->type)
+  {
+    case CONFIG_TYPE_INT64:
+      *value = setting->value.llval;
+      return(CONFIG_TRUE);
+
+    case CONFIG_TYPE_INT:
+      *value = (long long)(setting->value.ival);
+      return(CONFIG_TRUE);
+
+    case CONFIG_TYPE_FLOAT:
+      if((setting->config->flags & CONFIG_OPTION_AUTOCONVERT) != 0)
+      {
+        *value = (long long)(setting->value.fval);
+        return(CONFIG_TRUE);
+      }
+      else
+      { /* fall through */ }
+
+    default:
+      return(CONFIG_FALSE);
   }
 }
 
@@ -794,23 +853,9 @@ int config_setting_get_int(const config_setting_t *setting)
 
 long long config_setting_get_int64(const config_setting_t *setting)
 {
-  switch(setting->type)
-  {
-    case CONFIG_TYPE_INT64:
-      return(setting->value.llval);
-
-    case CONFIG_TYPE_INT:
-      return((long long)setting->value.ival);
-
-    case CONFIG_TYPE_FLOAT:
-      if((setting->config->flags & CONFIG_OPTION_AUTOCONVERT) != 0)
-        return((long long)(setting->value.fval));
-      else
-        /* fall through */;
-
-    default:
-      return(0);
-  }
+  long long value = 0;
+  __config_setting_get_int64(setting, &value);
+  return(value);
 }
 
 /* ------------------------------------------------------------------------- */
@@ -822,11 +867,7 @@ int config_setting_lookup_int(const config_setting_t *setting,
   if(! member)
     return(CONFIG_FALSE);
 
-  if(config_setting_type(member) != CONFIG_TYPE_INT)
-    return(CONFIG_FALSE);
-
-  *value = config_setting_get_int(member);
-  return(CONFIG_TRUE);
+  return(__config_setting_get_int(member, value));
 }
 
 /* ------------------------------------------------------------------------- */
@@ -838,11 +879,50 @@ int config_setting_lookup_int64(const config_setting_t *setting,
   if(! member)
     return(CONFIG_FALSE);
 
-  if(config_setting_type(member) != CONFIG_TYPE_INT64)
-    return(CONFIG_FALSE);
+  return(__config_setting_get_int64(member, value));
+}
 
-  *value = config_setting_get_int64(member);
-  return(CONFIG_TRUE);
+/* ------------------------------------------------------------------------- */
+
+static int __config_setting_get_float(const config_setting_t *setting,
+                                      double *value)
+{
+  switch(setting->type)
+  {
+    case CONFIG_TYPE_FLOAT:
+      *value = setting->value.fval;
+      return(CONFIG_TRUE);
+
+    case CONFIG_TYPE_INT:
+      if(config_get_auto_convert(setting->config))
+      {
+        *value = (double)(setting->value.ival);
+        return(CONFIG_TRUE);
+      }
+      else
+        return(CONFIG_FALSE);
+
+    case CONFIG_TYPE_INT64:
+      if(config_get_auto_convert(setting->config))
+      {
+        *value = (double)(setting->value.llval);
+        return(CONFIG_TRUE);
+      }
+      else
+      { /* fall through */ }
+
+    default:
+      return(CONFIG_FALSE);
+  }
+}
+
+/* ------------------------------------------------------------------------- */
+
+double config_setting_get_float(const config_setting_t *setting)
+{
+  double value = 0.0;
+  __config_setting_get_float(setting, &value);
+  return(value);
 }
 
 /* ------------------------------------------------------------------------- */
@@ -854,11 +934,7 @@ int config_setting_lookup_float(const config_setting_t *setting,
   if(! member)
     return(CONFIG_FALSE);
 
-  if(config_setting_type(member) != CONFIG_TYPE_FLOAT)
-    return(CONFIG_FALSE);
-
-  *value = config_setting_get_float(member);
-  return(CONFIG_TRUE);
+  return(__config_setting_get_float(member, value));
 }
 
 /* ------------------------------------------------------------------------- */
@@ -958,32 +1034,6 @@ int config_setting_set_int64(config_setting_t *setting, long long value)
 
 /* ------------------------------------------------------------------------- */
 
-double config_setting_get_float(const config_setting_t *setting)
-{
-  switch(setting->type)
-  {
-    case CONFIG_TYPE_FLOAT:
-      return(setting->value.fval);
-
-    case CONFIG_TYPE_INT:
-      if(config_get_auto_convert(setting->config))
-        return((double)(setting->value.ival));
-      else
-        return(0.0);
-
-    case CONFIG_TYPE_INT64:
-      if(config_get_auto_convert(setting->config))
-        return((double)(setting->value.llval));
-      else
-        return(0.0);
-
-    default:
-      return(0.0);
-  }
-}
-
-/* ------------------------------------------------------------------------- */
-
 int config_setting_set_float(config_setting_t *setting, double value)
 {
   switch(setting->type)
@@ -1078,9 +1128,10 @@ int config_setting_set_format(config_setting_t *setting, short format)
 
 /* ------------------------------------------------------------------------- */
 
-short config_setting_get_format(config_setting_t *setting)
+short config_setting_get_format(const config_setting_t *setting)
 {
-  return(setting->format);
+  return(setting->format != 0 ? setting->format
+         : setting->config->default_format);
 }
 
 /* ------------------------------------------------------------------------- */
@@ -1149,12 +1200,7 @@ int config_lookup_int(const config_t *config, const char *path,
   if(! s)
     return(CONFIG_FALSE);
 
-  if(config_setting_type(s) != CONFIG_TYPE_INT)
-    return(CONFIG_FALSE);
-
-  *value = config_setting_get_int(s);
-
-  return(CONFIG_TRUE);
+  return(__config_setting_get_int(s, value));
 }
 
 /* ------------------------------------------------------------------------- */
@@ -1166,12 +1212,7 @@ int config_lookup_int64(const config_t *config, const char *path,
   if(! s)
     return(CONFIG_FALSE);
 
-  if(config_setting_type(s) != CONFIG_TYPE_INT64)
-    return(CONFIG_FALSE);
-
-  *value = config_setting_get_int64(s);
-
-  return(CONFIG_TRUE);
+  return(__config_setting_get_int64(s, value));
 }
 
 /* ------------------------------------------------------------------------- */
@@ -1183,11 +1224,7 @@ int config_lookup_float(const config_t *config, const char *path,
   if(! s)
     return(CONFIG_FALSE);
 
-  if(config_setting_type(s) != CONFIG_TYPE_FLOAT)
-    return(CONFIG_FALSE);
-
-  *value = config_setting_get_float(s);
-  return(CONFIG_TRUE);
+  return(__config_setting_get_float(s, value));
 }
 
 /* ------------------------------------------------------------------------- */
@@ -1451,6 +1488,14 @@ void config_set_destructor(config_t *config, void (*destructor)(void *))
 
 /* ------------------------------------------------------------------------- */
 
+void config_set_include_dir(config_t *config, const char *include_dir)
+{
+  _delete(config->include_dir);
+  config->include_dir = strdup(include_dir);
+}
+
+/* ------------------------------------------------------------------------- */
+
 int config_setting_length(const config_setting_t *setting)
 {
   if((setting->type != CONFIG_TYPE_GROUP)
@@ -1558,7 +1603,7 @@ int config_setting_index(const config_setting_t *setting)
 
   list = setting->parent->value.list;
 
-  for(i = 0, found = list->elements; i < list->length; ++i, ++found)
+  for(i = 0, found = list->elements; i < (int)list->length; ++i, ++found)
   {
     if(*found == setting)
       return(i);
